package org.squin.benchmark.wodsim;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.zip.GZIPInputStream;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.hp.hpl.jena.graph.Factory;
import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.NodeFactory;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.sparql.graph.NodeTransform;
import org.apache.jena.atlas.lib.Sink;
import org.apache.jena.atlas.lib.SinkWrapper;
import org.apache.jena.riot.Lang;
import org.apache.jena.riot.RiotReader;
import org.apache.jena.riot.lang.SinkTriplesToGraph;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class WODSimServlet extends HttpServlet
{
	// general configuration
	static final public String DEFAULT_mainPrefixForURIs = "/WODSim/";
	static final public String DEFAULT_subPrefixForNIRs  = "id/";   // non-information resources
	static final public String DEFAULT_subPrefixForIRs   = "data/"; // information resources
	static final public long   DEFAULT_throttleSleepTime = 3000; // milliseconds

	static final public String mainPrefixForURIs = DEFAULT_mainPrefixForURIs;
	static final public String subPrefixForNIRs  = DEFAULT_subPrefixForNIRs;  // non-information resources
	static final public String subPrefixForIRs   = DEFAULT_subPrefixForIRs;   // information resources

	// members

	private Logger logger = LoggerFactory.getLogger( WODSimServlet.class );
	private File mainDataDir;
	private String categoryOfDataFiles;
	private long throttleSleepTime;


	// implementation of the HttpServlet interface

	public void init ( ServletConfig config ) throws ServletException
	{
		logger.debug( "Initializing {} ...", toString() );

		super.init( config );

		String mainDataDirName = config.getInitParameter( "DataFilesMainDir" );
		if ( mainDataDirName == null ) {
			throw new ServletException( "Initialization error (parameter 'DataFilesMainDir' not set)." );
		}

		mainDataDir = new File ( mainDataDirName );
		if ( ! mainDataDir.exists() ) {
			throw new ServletException( "Initialization error (parameter 'DataFilesMainDir' refers to a non-existing directory)." );
		}
		if ( ! mainDataDir.isDirectory() ) {
			throw new ServletException( "Initialization error (parameter 'DataFilesMainDir' refers to a file but not to a directory)." );
		}

		categoryOfDataFiles = config.getInitParameter( "DataFilesCategorySubDir" );
		if ( categoryOfDataFiles == null ) {
			throw new ServletException( "Initialization error (parameter 'DataFilesCategorySubDir' not set)." );
		}

		File tmp = new File ( mainDataDir, categoryOfDataFiles );
		if ( ! tmp.exists() ) {
			throw new ServletException( "Initialization error (parameter 'DataFilesCategorySubDir' refers to a non-existing sub-directory within 'DataFilesMainDir')." );
		}
		if ( ! tmp.isDirectory() ) {
			throw new ServletException( "Initialization error (parameter 'DataFilesCategorySubDir' refers to a file within 'DataFilesMainDir', but not to a directory)." );
		}

		String throttleSleepTimeString = config.getInitParameter( "SleepTimeForResponseTimeThrottling" );
		try {
			throttleSleepTime = ( throttleSleepTimeString != null ) ? Long.parseLong( throttleSleepTimeString ) : DEFAULT_throttleSleepTime;
		}
		catch ( NumberFormatException e ) {
			throw new ServletException( "Initialization error (parsing parameter 'SleepTimeForResponseTimeThrottling' into a long integer failed: " + e.getMessage() + ")." );
		}

		logger.debug( "... initialization of {} completed", toString() );
	}

	protected void doGet ( HttpServletRequest req, HttpServletResponse resp )
	{
		String reqURI = req.getRequestURI();
		if ( req.getQueryString() != null ) {
			reqURI += "?";
			reqURI += req.getQueryString();
		}

		logger.info( "Start processing request {} for {}.", req.hashCode(), reqURI );

		if ( ! reqURI.startsWith(mainPrefixForURIs) )
		{
			logger.error( "The request {} for URI <{}> does not start with the expected prefix <{}>.", new Object[] {req.hashCode(),reqURI,mainPrefixForURIs} );

			try {
				resp.sendError( HttpServletResponse.SC_INTERNAL_SERVER_ERROR );
			} catch ( IOException e ) {
				logger.error( "Sending the error response for request " + req.hashCode() + " caused a " + e.getClass().getName() + ": " + e.getMessage(), e );
			}

			return;
		}

		reqURI = reqURI.substring( mainPrefixForURIs.length() );

		if ( reqURI.startsWith(subPrefixForNIRs) ) {
			handleRequestForNIR( reqURI.substring(subPrefixForNIRs.length()), req, resp );
		}
		else if ( reqURI.startsWith(subPrefixForIRs) ) {
			handleRequestForIR( reqURI.substring(subPrefixForIRs.length()), req, resp );
		}
		else {
			handleUnexpectedRequest( req, resp );
		}

		logger.info( "Finished processing request {}.", req.hashCode() );
	}


	// helper methods

	protected void handleUnexpectedRequest ( HttpServletRequest req, HttpServletResponse resp )
	{
		respondWith404( req, resp );
	}

	protected void handleRequestForNIR ( String uri, HttpServletRequest req, HttpServletResponse resp )
	{
		if ( ! uri.startsWith("http://") ) {
			respondWith404( req, resp );
			return;
		}

		String relPath = uri.substring( 7 );
		String redirectLocationAtWODSim = req.getScheme() + "://" +
		                                  req.getServerName() + ":" +
		                                  req.getServerPort() + mainPrefixForURIs +
		                                  subPrefixForIRs + categoryOfDataFiles +
		                                  relPath;
		try {
// 			resp.sendRedirect( redirectLocationAtWODSim );
			resp.setHeader( "Location", redirectLocationAtWODSim );
			resp.setStatus( HttpServletResponse.SC_SEE_OTHER );
			resp.flushBuffer();
			logger.info( "Answered request {} with a redirect to <{}>.", req.hashCode(), redirectLocationAtWODSim );
		} catch ( IOException e ) {
			logger.error( "Sending a redirect for request " + req.hashCode() + " caused a " + e.getClass().getName() + ": " + e.getMessage(), e );
		}
	}

	protected void handleRequestForIR ( String relPath, HttpServletRequest req, HttpServletResponse resp )
	{
		File dataFile = getDataFile( relPath );
		if ( dataFile == null ) {
			respondWith404( req, resp );
			return;
		}

		InputStream inStream = null;
		try {
			inStream = new GZIPInputStream( new FileInputStream(dataFile) );
		}
		catch ( IOException e )
		{
			logger.error( "Getting the input stream for file '" + dataFile.getAbsolutePath() + "' caused a " + e.getClass().getName() + ": " + e.getMessage(), e );

			try {
				resp.sendError( HttpServletResponse.SC_INTERNAL_SERVER_ERROR );
			} catch ( IOException e2 ) {
				logger.error( "Sending the error response for request " + req.hashCode() + " caused a " + e2.getClass().getName() + ": " + e2.getMessage(), e2 );
			}
		}

		Graph outputGraph;
		try {
			outputGraph = Factory.createGraphMem();
// ATTENTION: For some reason the SinkTriplesToGraph constructor takes a
// boolean as first parameter but does not make use of it. 
// 			SinkTriplesToGraph wrappedSink = new SinkTriplesToGraph( outputGraph );
SinkTriplesToGraph wrappedSink = new SinkTriplesToGraph( false, outputGraph );

			String prefixForRewritingURIs = req.getScheme() + "://" +
			                                req.getServerName() + ":" +
			                                req.getServerPort() + mainPrefixForURIs +
			                                subPrefixForNIRs;
			SinkTripleNodeTransform sink = new SinkTripleNodeTransform( wrappedSink, new RewriteURIs(prefixForRewritingURIs) );

			String baseURI = null;
			RiotReader.parseTriples( inStream, Lang.TURTLE, baseURI, sink );
			sink.close();
		}
		catch ( Exception e ) {
			logger.error( "Parsing the input stream for request " + req.hashCode() + " caused a " + e.getClass().getName() + ": " + e.getMessage(), e );
			outputGraph = null;
		}

		try {
			inStream.close();
		}
		catch ( IOException e ) {
			logger.error( "Closing the input stream for request " + req.hashCode() + " caused a " + e.getClass().getName() + ": " + e.getMessage(), e );
			outputGraph = null;
		}

		if ( outputGraph == null )
		{
			try {
				resp.sendError( HttpServletResponse.SC_INTERNAL_SERVER_ERROR );
			} catch ( IOException e ) {
				logger.error( "Sending the error response for request " + req.hashCode() + " caused a " + e.getClass().getName() + ": " + e.getMessage(), e );
			}

			return;
		}

		OutputStream outStream;
		try {
			outStream = resp.getOutputStream();
		}
		catch ( IOException e )
		{
			logger.error( "Getting the response output stream for request " + req.hashCode() + " caused a " + e.getClass().getName() + ": " + e.getMessage(), e );

			try {
				resp.sendError( HttpServletResponse.SC_INTERNAL_SERVER_ERROR );
			} catch ( IOException e2 ) {
				logger.error( "Sending the error response for request " + req.hashCode() + " caused a " + e2.getClass().getName() + ": " + e2.getMessage(), e2 );
			}

			return;
		}

		try {
			Thread.sleep( throttleSleepTime ); // throttle the response time
		} catch ( InterruptedException e ) {
			// ignore
		}

		resp.setContentType( Lang.TURTLE.getContentType().getContentType() );
// 		resp.setContentLength( (int) dataFile.length() );
		try {
			ModelFactory.createModelForGraph( outputGraph ).write( outStream, "TURTLE" );
		}
		catch ( Exception e )
		{
			logger.error( "Writing to the response stream for request {} caused a {}: {}", new Object[] {req.hashCode(),e.getClass().getName(),e.getMessage()} );
			try {
				resp.sendError( HttpServletResponse.SC_INTERNAL_SERVER_ERROR );
			} catch ( IOException e2 ) {
				logger.error( "Sending an error response for request " + req.hashCode() + " caused a " + e2.getClass().getName() + ": " + e2.getMessage(), e2 );
			}
		}

		// finish
		try {
			outStream.flush();
			resp.flushBuffer();
			outStream.close();
		}
		catch ( IOException e ) {
			logger.error( "Flushing the response buffer for request " + req.hashCode() + " caused a " + e.getClass().getName() + ": " + e.getMessage(), e );
		}
	}

	protected File getDataFile ( String relPath )
	{
		if ( relPath.contains("#") || relPath.contains("?") ) {
			return null;
		}

		File f = new File( mainDataDir, relPath );
		return ( f.exists() && f.isFile() && f.canRead() ) ? f : null;
	}

	protected void respondWith404 ( HttpServletRequest req, HttpServletResponse resp )
	{
		try {
			resp.sendError( HttpServletResponse.SC_NOT_FOUND );
			logger.info( "Answered request {} with a 404.", req.hashCode() );
		} catch ( IOException e ) {
			logger.error( "Sending a 404 response for request " + req.hashCode() + " caused a " + e.getClass().getName() + ": " + e.getMessage(), e );
		}
	}

	static public void copy ( InputStream in, OutputStream out ) throws IOException
	{
		byte[] buffer = new byte[ 1024 ];
		int count = 0;
		int n = 0;
		while ( (n = in.read(buffer)) != -1 ) {
			out.write( buffer, 0, n );
		}
	}


	static class RewriteURIs implements NodeTransform
	{
		final public String prefix;

		public RewriteURIs ( String prefix ) { this.prefix = prefix; }

		public Node convert ( Node node )
		{
			if ( ! node.isURI() ) {
				return node;
			}

			String uri = node.getURI();
			if (    uri.startsWith(com.hp.hpl.jena.vocabulary.RDF.getURI())
			     || uri.startsWith(com.hp.hpl.jena.vocabulary.RDFS.getURI())
			     || uri.startsWith(com.hp.hpl.jena.vocabulary.OWL.getURI()) ) {
				return node;
			}

			return NodeFactory.createURI( prefix + uri );
		}
	}

	static class SinkTripleNodeTransform extends SinkWrapper<Triple>
	{
		final protected NodeTransform transformer;
		public SinkTripleNodeTransform ( Sink<Triple> wrappedSink, NodeTransform transformer ) { super(wrappedSink); this.transformer = transformer; }

		@Override
    	public void send ( Triple t )
    	{
			Node oldS = t.getSubject();
			Node oldP = t.getPredicate();
			Node oldO = t.getObject();
        
			Node newS = transformer.convert( oldS );
			Node newP = transformer.convert( oldP );
			Node newO = transformer.convert( oldO );

			if ( oldS != newS || oldP != newP || oldO != newO ) { 
				super.send( new Triple(newS,newP,newO) );
			} else {
				super.send( t );
			}
		}
	}

}